/*
 * Copyright (c) 2019, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * WSO2 Inc. licenses this file to you under the Apache License,
 * Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package io.siddhi.distribution.test.framework.containers;

import com.google.common.collect.ImmutableMap;
import io.siddhi.distribution.test.framework.KafkaContainer;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.apache.kafka.common.serialization.StringSerializer;
import org.rnorth.ducttape.unreliables.Unreliables;
import org.testcontainers.containers.GenericContainer;
import org.testcontainers.containers.Network;
import org.testng.annotations.Test;

import java.util.Arrays;
import java.util.UUID;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.AssertionsForClassTypes.tuple;

/**
 * Kafka Container test class.
 */
public class KafkaContainerTest {

    @Test
    public void testUsage() throws Exception {
        try (KafkaContainer kafka = new KafkaContainer()) {
            kafka.start();
            testKafkaFunctionality(kafka.getBootstrapServers());
        }
    }

    @Test(dependsOnMethods = {"testUsage"})
    public void testExternalZookeeperWithKafkaNetwork() throws Exception {
        try (
                KafkaContainer kafka = new KafkaContainer()
                        .withExternalZookeeper("zookeeper:2181");
                GenericContainer zookeeper = new GenericContainer("confluentinc/cp-zookeeper:4.0.0")
                        .withNetwork(kafka.getNetwork())
                        .withNetworkAliases("zookeeper")
                        .withEnv("ZOOKEEPER_CLIENT_PORT", "2181");
        ) {
            Stream.of(kafka, zookeeper).parallel().forEach(GenericContainer::start);
            testKafkaFunctionality(kafka.getBootstrapServers());
        }
    }

    @Test(dependsOnMethods = {"testExternalZookeeperWithKafkaNetwork"})
    public void testExternalZookeeperWithExternalNetwork() throws Exception {
        Network network = Network.newNetwork();
        GenericContainer zookeeper = null;
        KafkaContainer kafka = null;
        try {
            zookeeper = new GenericContainer("confluentinc/cp-zookeeper:4.0.0")
                    .withNetwork(network)
                    .withNetworkAliases("zookeeper")
                    .withEnv("ZOOKEEPER_CLIENT_PORT", "2181");
            zookeeper.start();
            Thread.sleep(5000);
            kafka = new KafkaContainer()
                    .withNetwork(network)
                    .withExternalZookeeper("zookeeper:2181");
            kafka.start();
            testKafkaFunctionality(kafka.getBootstrapServers());
        } finally {
            network.close();
            if (zookeeper != null) {
                zookeeper.stop();
            }
            if (kafka != null) {
                kafka.stop();
            }
        }
    }

    private void testKafkaFunctionality(String bootstrapServers) throws Exception {
        try (
                KafkaProducer<String, String> producer = new KafkaProducer<>(
                        ImmutableMap.of(
                                ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers,
                                ProducerConfig.CLIENT_ID_CONFIG, UUID.randomUUID().toString()
                        ),
                        new StringSerializer(),
                        new StringSerializer()
                );
                KafkaConsumer<String, String> consumer = new KafkaConsumer<>(
                        ImmutableMap.of(
                                ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers,
                                ConsumerConfig.GROUP_ID_CONFIG, "tc-" + UUID.randomUUID(),
                                ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest"
                        ),
                        new StringDeserializer(),
                        new StringDeserializer()
                );
        ) {
            String topicName = "messages";
            consumer.subscribe(Arrays.asList(topicName));
            producer.send(new ProducerRecord<>(topicName, "Siddhi", "rulezzz")).get();
            Unreliables.retryUntilTrue(10, TimeUnit.SECONDS, () -> {
                ConsumerRecords<String, String> records = consumer.poll(100);
                if (records.isEmpty()) {
                    return false;
                }
                assertThat(records).hasSize(1).extracting(ConsumerRecord::topic,
                        ConsumerRecord::key, ConsumerRecord::value)
                        .containsExactly(tuple(topicName, "Siddhi", "rulezzz"));
                return true;
            });
            consumer.unsubscribe();
        }
    }
}
